#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <string>
#include <vector>

#include <sys/socket.h>
#include <sys/stat.h>
#include <poll.h>

#include "nesvideos-piece.hh"

static int ForceFullRead(unsigned char* buf, unsigned size)
{
    int got = 0;
    while(size > 0)
    {
        int res = read(0, buf, (size_t)size);
        if(res <= 0) { perror("read"); return got; }
        buf += res;
        size -= res;
        got += res;
    }
    return got;
}
static int ForceFullWrite(int fd, const unsigned char* buf, unsigned size)
{
    int got = 0;
    while(size > 0)
    {
        int res = write(fd, buf, (size_t)size);
        if(res <= 0) { perror("write"); return got; }
        buf += res;
        size -= res;
        got += res;
    }
    return got;
}


static FILE* resampler = NULL;
static void CloseResampler()
{
    if(resampler) { pclose(resampler); resampler = NULL; }
}
static void OpenResampler(unsigned InputRate, unsigned OutputRate,
                          unsigned Channels,
                          unsigned Bitness)
{
    int pip[2];
    socketpair(AF_UNIX, SOCK_STREAM, 0, pip);
    if(!fork())
    {
        char Buf[512];
        
        fprintf(stderr,
            "\n\nStarting a resampler from %u Hz to %u Hz\n\n",
            InputRate, OutputRate);
        
        std::string inputspec;
        sprintf(Buf, "r%u", InputRate);
        inputspec += Buf;

        std::string outputspec;
        if(Channels == 1) outputspec += 'm';
        if(Bitness  == 8) outputspec += 'b';
        sprintf(Buf, "r%u", OutputRate);
        outputspec += Buf;
        
        int p0 = dup(dup(pip[0]));
        dup2(p0, 0);
        dup2(p0, 1);
        for(int n=3; n<511; ++n) close(n);
        
        std::string filter = "lanczos";
        
        execlp("resample",
               "resample",
               inputspec.c_str(),
               outputspec.c_str(),
               ("-f" + filter).c_str(),
               NULL);
        
        perror("execlp(\"resample\")");
        fprintf(stderr,
            "The resample program is required for stream resampling in this\n"
            "program. Get it at: http://bisqwit.iki.fi/source/streamixer.html\n"
        );
        _exit(-1);
    }
    close(pip[0]);
     
    resampler = fdopen(pip[1], "r+");
}   

int main(int argc, const char*const* argv)
{
    fprintf(stderr, "Received a connection!\n");
    argc=argc;
    
    NESVideoSetVideoCmd(argv[1]);
    LoggingEnabled = 2;
    
    unsigned InRate = 0;
    const unsigned OutRate = 48000;
    unsigned Bits = 16;
    unsigned Chans = 2;
    
    std::vector<unsigned char> ResampleBuf;
    
    for(;;)
    {
        if(resampler)
        {
            struct pollfd p[2];
            p[0].fd     = 0; 
            p[0].events = POLLIN;
            p[1].fd     = fileno(resampler);
            p[1].events = POLLIN;
            
            if(!ResampleBuf.empty())
            {
                p[1].events |= POLLOUT;
            }
            
            poll(p, 2, -1);
            
            if(p[1].revents & POLLERR)
            {
                CloseResampler();
                continue;
            }
            
            if(p[1].revents & POLLOUT)
            {
                int res = write(fileno(resampler), &ResampleBuf[0], ResampleBuf.size());
                if(res > 0)
                {
                    ResampleBuf.erase(ResampleBuf.begin(), ResampleBuf.begin() + res);
                }
            }
            
            if(p[1].revents & POLLIN)
            {
                char Buf[65536];
                int res = read(fileno(resampler), Buf, sizeof(Buf));
                if(res > 0)
                {
                    int Samples = res / (Bits/8) / Chans;
                    NESVideoLoggingAudio(Buf, OutRate,Bits,Chans, Samples);
                }
            }
            
            if(p[0].revents & POLLERR)
            {
                break;
            }
            
            if(!(p[0].revents & POLLIN)) continue;
        }
    
        unsigned char Header[16] = "?";
        if(ForceFullRead(Header, 16) <= 0) break;
        
        switch(Header[0])
        {
            case 'V':
            {
                unsigned w = *(unsigned*)&Header[4];
                unsigned h = *(unsigned*)&Header[8];
                unsigned f = *(unsigned*)&Header[12];
                unsigned bytes = w*h*3;

                unsigned char Buf[bytes], Buf2[bytes];
                ForceFullRead(Buf, bytes);
                
#if 1 /* flip? */
                // flip the image vertically.
                for(unsigned a=0; a<h; ++a)
                    memcpy(&Buf2[a * (w*3)],
                           &Buf[ (h-a-1) * (w*3)],
                           (size_t)(w*3));
                
                NESVideoLoggingVideo(Buf2, w,h,f);
#else /* no flip */
                NESVideoLoggingVideo(Buf, w,h,f);
#endif
                break;
            }
            case 'A':
            {
                unsigned rate = *(unsigned*)&Header[4];
                unsigned bits = *(unsigned short*)&Header[8];
                unsigned chans = *(unsigned short*)&Header[10];
                unsigned samples = *(unsigned*)&Header[12];
                unsigned bytes = samples*(bits/8)*chans;
                
                unsigned char Buf[bytes];
                ForceFullRead(Buf, bytes);
                
                /* If the input rate has changed, try to flush whatever
                 * audio we have got so far, and then start a new resampler.
                 */
                if(rate != InRate && resampler)
                {
                    ForceFullWrite(fileno(resampler),
                                   &ResampleBuf[0], ResampleBuf.size());
                    ResampleBuf.clear();
                    
                flush_resa_more: ;
                    struct pollfd p = { fileno(resampler), POLLIN, 0 };
                    poll(&p, 1, 2500); // wait 2.5 seconds
                    if(p.revents & POLLIN)
                    {
                        char Buf[65536];
                        int res = read(fileno(resampler), Buf, sizeof(Buf));
                        if(res > 0)
                        {
                            int Samples = res / (Bits/8) / Chans;
                            NESVideoLoggingAudio(Buf, OutRate,Bits,Chans, Samples);
                        }
                        goto flush_resa_more;
                    }
                    CloseResampler();
                }
                
                if(!resampler)
                {
                    OpenResampler(rate, OutRate, chans, bits);
                }
                
                InRate = rate;
                Chans = chans;
                Bits  = bits;
                
                ResampleBuf.insert(ResampleBuf.end(), Buf, Buf+bytes);
                break;
            }
        }
    }
    
    return 0;
}
